//============================================================================
// Bare-Bones Tablet Interface for C++
//
//  Implementation file
//
//  Author:        William Baxter (baxter@cs.unc.edu)
//  Last Modified: Aug 2002
//============================================================================
//  Copyright 2002 
//        William Baxter and
//        The University of North Carolina at Chapel Hill
// 
//  Permission to use, copy, modify, distribute and sell this software
//  and its documentation for any purpose is hereby granted without
//  fee, provided that the above copyright notice appear in all copies
//  and that both that copyright notice and this permission notice
//  appear in supporting documentation.  Binaries may be compiled with
//  this software without any royalties or restrictions.
// 
//  The University of North Carolina at Chapel Hill makes no
//  representations about the suitability of this software for any
//  purpose. It is provided "as is" without express or implied
//  warranty.
//============================================================================

#if defined (_WIN32)
# define WIN32_LEAN_AND_MEAN
# include <windows.h>
#endif

/* Because we'll look up functions from the DLL, don't prototype them */
#define NOWTFUNCTIONS
#include <wintab.h>

#include <sys/timeb.h> // for waitForNextEvent
#include <vector>
#include <string>
#include <cassert>
using std::vector;
using std::string;

#if 0
// mouse-like packet
#define PACKETNAME  MOUSE
#define MOUSEPACKETDATA PK_X | PK_Y | PK_BUTTONS
#define MOUSEPACKETMODE 0 /* absolute mode */
#include <pktdef.h>
#undef PACKETNAME

// basic wacom-like packet
#define PACKETNAME  BASIC
#define BASICPACKETDATA PK_X | PK_Y | PK_BUTTONS | PRESSURE
#define BASICPACKETMODE 0 /* absolute mode */
#include <pktdef.h>
#undef PACKETNAME
#endif /* 0 */

// We want it all!
#define PACKETNAME FULL
#define FULLPACKETDATA	\
     (   PK_X | PK_Y | PK_Z | PK_BUTTONS \
       | PK_CONTEXT \
       | PK_STATUS  \
       | PK_TIME    \
       | PK_CHANGED \
       /*| PK_SERIAL_NUMBER *//* that's a packet SN, not pen SN*/\
       | PK_CURSOR \
       | PK_NORMAL_PRESSURE \
       | PK_TANGENT_PRESSURE /* the scroll wheel on wacom airbrush */\
       | PK_ORIENTATION \
       | PK_ROTATION \
 )
#define FULLPACKETMODE	0 /* ABSOLUTE MODE */
#define FULLPACKETTILT	PKEXT_ABSOLUTE
#include <pktdef.h>
#undef PACKETNAME

#include "bbtablet.h"
#include "bbtabwin.h"
#include <math.h>
#include <stdio.h>

#ifndef M_PI
static const double M_PI = 3.1415926535897932;
#endif
static const float M_PIf = float(M_PI);

// bbTabletEvent flags
enum {
  BBTEV_ORIENT_OK=1
};

inline float fix2float(FIX32 arg)
{
  return arg * (1.0f/65536.0f);
}

#if defined(DEBUG) || defined(_DEBUG)
void DebugMsg(const char *fmt, ...)
{
  va_list args;
  va_start(args, fmt);
  vfprintf(stderr, fmt, args);
  va_end(args);
}
#else 
inline void DebugMsg(const char *fmt, ...) { }
#endif


namespace {

void convertOrientation2AngleAxis(const ORIENTATION &orient,
    float scale2rads,
    float *angle,
    float axis[3])
  {
    bool flip = orient.orAltitude < 0; 

    float  az = orient.orAzimuth  * scale2rads;
    float alt = orient.orAltitude * scale2rads;
    float  tw = orient.orTwist    * scale2rads;

    //   this is sort of like an euler zyz angles representation, 
    //   but a bit different since altitude depends on whether the pen
    //   is upside down or not, but azimuth doesn't care.
    //   Also we want z up to be (0,0,0) in angles, not horizontal.

    //  Remap altitude: 0 is straight up, pi is straight upside down
    alt = M_PIf*0.5f - alt;
    //  Remap azimuth: 0 is +x axis, and increases CCW, decreases CW
    if (az > M_PI*1.5f) {
      az = M_PIf*2.5f - az;
    }
    else {
      az = M_PIf*0.5f - az;
    }


    float sa = (float)sin(alt);
    float ca = (float)cos(alt);
    float st = (float)sin(tw);
    float ct = (float)cos(tw);
    float sz = (float)sin(az);
    float cz = (float)cos(az);
    if (!flip) {
      // In Matlab terms, we have the following:
      // Raz_z = [ cz -sz  0 ; sz   cz    0 ;  0   0   1 ];
      // Ral_y = [ ca  0  sa ;  0    1    0 ; -sa  0  ca ];
      // Rtw_z = [ ct -st  0 ; st   ct    0 ;  0   0   1 ];
      // R = Raz_z * Ral_y * Raz_z.' * Rtw_z;
      // That's where the trace, and rXYmrYX terms below come from

      float trace = ct*(ca+1.0f)+ca;

      float cosAngle = 0.5f*(trace - 1.0f);
      *angle = float( acos(cosAngle) );
      if (*angle == 0) {
        axis[0] = 0;
        axis[1] = 0;
        axis[2] = 1.0f;
      }
      else {
        float r32mr23 = sa*( st*cz - ct*sz - sz);
        float r21mr12 = (ca+1)*st;
        float r13mr31 = sa*( cz + ct*cz + st*sz);

        float coef = 0.5f / (float)sin(*angle);
        axis[0] = coef * (r32mr23);
        axis[1] = coef * (r13mr31);
        axis[2] = coef * (r21mr12);
      }
    }
    else { // FLIP IT!
      // Flipped case is handled differently so that there won't be a singularity
      // when the pen is exactly upside down.
      // In Matlab terms, we have the following:
      // Raz_z = [ cz -sz  0 ; sz   cz    0 ;  0   0   1 ];
      // Ral_y = [ ca  0  sa ;  0    1    0 ; -sa  0  ca ];
      // Rtw_z = [ ct -st  0 ; st   ct    0 ;  0   0   1 ];
      // roty180 = [ -1 0 0; 0 1 0; 0 0 -1 ];

      // R = roty180 * Raz_z * Ral_y * Raz_z.' * Rtw_z;
      // We also modify some of the angles:
      //   * use -az instead of az.
      //   * use 180-alt instead of alt.
      // That's where the trace, and rXYmrYX terms below come from

      sz = -sz; // az  => -az
      ca = -ca; // alt => 180 - alt

      float szcz2 = 2*sz*cz;
      float cz2msz2 = cz*cz - sz*sz;
      float va = (1.0f-ca);
      float trace = va * ( szcz2*st + cz2msz2*ct ) - ca;

      float cosAngle = 0.5f*(trace - 1.0f);
      *angle = float( acos(cosAngle) );
      if (*angle == 0) {
        axis[0] = 0;
        axis[1] = 0;
        axis[2] = 1.0f;
      }
      else {
        float r32mr23 = sa * (sz*ct - st*cz - sz);
        float r21mr12 =  va*( -szcz2*ct + cz2msz2*st ) ;
        float r13mr31 = -sa*(cz + ct*cz + sz*st);

        float coef = 0.5f / float( sin(*angle) );
        axis[0] = coef * (r32mr23);
        axis[1] = coef * (r13mr31);
        axis[2] = coef * (r21mr12);
      }
    }

  }
} // namespace

//----------------------------------------------------------------------------
/**
 * @class _bbTabletCursor
 */
class _bbTabletCursor
{
public:
  // enum { PUCK, STYLUS, ERASER, PUCK2, STYLUS2, ERASER2 };
  string name;
  int nButtons;
  vector<string> buttonNames;
};

class _bbTabletDeviceInternal
{
public:

  _bbTabletDeviceInternal()
    : tabCtx(0), hwnd(0), lostPackets(0), throwAwayOldPackets(false)
  { }

  HWND hwnd;
  HCTX tabCtx;
  float AspectRatio;
  float scaleX;
  float scaleY;
  float scaleZ;
  float scalePress;
  float scaleTPress;
  float scaleAng;
  long lostPackets;
  bool throwAwayOldPackets;
  vector<_bbTabletCursor> cursors;

  void _convertOrientation2AngleAxis(
    const ORIENTATION &orient, float *angle, float axis[3])
  {
    ::convertOrientation2AngleAxis(orient,scaleAng,angle,axis);
  }
  friend class bbTabletEvent;
};

//----------------------------------------------------------------------------

bbTabletDevice &bbTabletDevice::getInstance()
{
  static bbTabletDevice td_singleton;
  return td_singleton;
}

bbTabletDevice::bbTabletDevice()
{
  m_data = new _bbTabletDeviceInternal();
}


bool bbTabletDevice::initTablet( void *sys_win, int mode )
{
  if (initWintabDLL()) {
    DebugMsg("You've got no Wintab32.dll\n");
    return false;
  }

  // Find out about our tablet
  if (!pfWTInfo(0,0,NULL)) {
    // no tablet or no tablet support
    DebugMsg("No tablet or no tablet support\n");
    return false;
  }

  m_data->hwnd = (HWND)sys_win;

  TCHAR strbuf[1024];

  /////////////////////////////////
  // Tablet info
  int nCursors = 0;
  {
    UINT ret = pfWTInfo(WTI_INTERFACE, IFC_WINTABID, strbuf);
    DebugMsg("Tablet interface provider: %s\n", LPCSTR(strbuf));
    UINT ntabs;
    UINT val;
    ret = pfWTInfo(WTI_INTERFACE, IFC_NDEVICES, &ntabs);
    DebugMsg("Supported # of devices: %d\n", ntabs);
    if (ret = pfWTInfo(WTI_INTERFACE, IFC_NCURSORS, &nCursors))
      DebugMsg("Supported # of cursors: %d\n", nCursors);
    
    ret = pfWTInfo(WTI_INTERFACE, IFC_NCONTEXTS, &val);
    DebugMsg("Supported # of contexts: %d\n", val);
    
    int tabnum=0;
    while (0 != (ret = pfWTInfo(WTI_DEVICES+tabnum, DVC_NAME, strbuf)) ) {
      
      DebugMsg("Tablet %d name: \"%s\"\n", tabnum, LPCSTR(strbuf));
      
      UINT hwFlags=0;
      ret = pfWTInfo(WTI_DEVICES+tabnum, DVC_HARDWARE, &hwFlags);
      if (hwFlags & HWC_INTEGRATED) 
        DebugMsg("   Has integrated screen.\n");
      if (hwFlags & HWC_TOUCH)
        DebugMsg("   Is touch-screen variety.\n");
      if (hwFlags & HWC_HARDPROX)
        DebugMsg("   Reports enter/leave events.\n");
      if (hwFlags & HWC_PHYSID_CURSORS)
        DebugMsg("   Uses unique cursor IDs.\n");
      tabnum++;
    }
    tabnum--;
    if (tabnum < 0) {
      DebugMsg("No tablet devices found\n");
    }
  }


  /////////////////////////////////
  // See what kind of cursors there are
  UINT ret;
  int i;
  for(i=0; 0 != (ret=pfWTInfo( WTI_CURSORS + i, CSR_NAME, strbuf )); i++) {
    //pfWTInfo( WTI_CURSORS + i, CSR_NAME, strbuf );
    DebugMsg("Cursor %d: \"%s\"\n", i, strbuf);
    
    m_data->cursors.push_back(_bbTabletCursor());
    m_data->cursors[i].name = strbuf;

    {
      // browse WinTab's many info items to discover pressure handling.
      int wActiveCsr = i;
      BYTE logBtns[32];
      UINT size;

      BOOL bActive;
      ret = pfWTInfo(WTI_CURSORS + wActiveCsr, CSR_ACTIVE, &bActive);
      if (ret)
        DebugMsg("  %s.\n", bActive?"Active":"Not active");

      WTPKT pktData;
      ret = pfWTInfo(WTI_CURSORS + wActiveCsr, CSR_PKTDATA, &pktData);
      if (bActive && ret) {
        DebugMsg("  Supports the following: ");
        if (pktData & PK_X) { DebugMsg("x "); }
        if (pktData & PK_Y) { DebugMsg("y "); }
        if (pktData & PK_Z) { DebugMsg("z "); }
        if (pktData & PK_BUTTONS) { DebugMsg("btns "); }
        if (pktData & PK_CONTEXT) { DebugMsg("ctx "); }
        if (pktData & PK_STATUS) { DebugMsg("stat "); }
        if (pktData & PK_TIME) { DebugMsg("tm "); }
        if (pktData & PK_CHANGED) { DebugMsg("chng "); }
        if (pktData & PK_SERIAL_NUMBER) { DebugMsg("sn "); }
        if (pktData & PK_CURSOR) { DebugMsg("csr "); }
        if (pktData & PK_NORMAL_PRESSURE) { DebugMsg("np "); }
        if (pktData & PK_TANGENT_PRESSURE) { DebugMsg("tp "); }
        if (pktData & PK_ORIENTATION) { DebugMsg("or "); }
        if (pktData & PK_ROTATION) { DebugMsg("rot "); }
        DebugMsg("\n");
      }

      DWORD physID = 0;
      ret = pfWTInfo(WTI_CURSORS + wActiveCsr, CSR_PHYSID, &physID);
      if (ret)
        DebugMsg("  physical id: %d\n", physID);

      BYTE nButtons = 0;
      pfWTInfo(WTI_CURSORS+wActiveCsr, CSR_BUTTONS, &nButtons);
      DebugMsg("  has %d buttons\n", nButtons);
      m_data->cursors[i].nButtons = nButtons;


      ret = pfWTInfo(WTI_CURSORS+wActiveCsr, CSR_BTNNAMES, strbuf);
      if (ret && nButtons>0) {
         DebugMsg("  Button names:\n");
         char *p = strbuf;
         for (int j=0; j<nButtons; j++) {
           DebugMsg("  %d - \"%s\"\n", j, p);
           m_data->cursors[i].buttonNames.push_back( string(p) );
           p+=strlen(p)+1;
         }
      }

      // discover the LOGICAL button generated by the pressure channel
      // get the PHYSICAL button from the cursor category and run it 
      // through that cursor's button map (usually the identity map). 
      BYTE wPrsBtn = (BYTE)-1;
      ret = pfWTInfo(WTI_CURSORS + wActiveCsr, CSR_NPBUTTON, &wPrsBtn);
      size = pfWTInfo(WTI_CURSORS + wActiveCsr, CSR_BUTTONMAP, &logBtns);
      if ((UINT)wPrsBtn < size)
        wPrsBtn = logBtns[wPrsBtn];

      // get the current context for its device variable. 
      // pfWTGet(hTab, &lc);

      // get the button marks (up & down generation thresholds) 
      // and calculate what pressure range we get when pressure-button is down. 
      
      /*
      AXIS np;
      UINT btnMarks[2];
      btnMarks[1] = 0; // default if info item not present. 
      ret = pfWTInfo(WTI_CURSORS + wActiveCsr, CSR_NPBTNMARKS, btnMarks);
      UINT prsYesBtnOrg = btnMarks[1];
      UINT prsYesBtnExt = (UINT)np.axMax - btnMarks[1];
      */
    }
  }
  return setPointerMode(mode);
}


bool bbTabletDevice::setPointerMode( int mode )
{
  // switch between SYSTEM_POINTER and SEPARATE_POINTER modes
  if (m_data && m_data->tabCtx && pointerMode() != mode) {
    // close current context
    pfWTClose(m_data->tabCtx);
    m_data->tabCtx = 0;
  }
  
  // create a new context with the new mode
  LOGCONTEXT lcMine;

  // UINT contextUsed = WTI_DEFSYSCTX;
  UINT contextUsed = (mode==SYSTEM_POINTER) ? WTI_DEFSYSCTX : WTI_DEFCONTEXT;
  //UINT contextUsed = (mode==SYSTEM_POINTER) ? WTI_DSCTXS : WTI_DDCTXS;
  // get default context
  // UINT ret = WTInfo(WTI_DEFCONTEXT, 0, &lcMine);
  UINT ret = pfWTInfo(contextUsed, 0, &lcMine);
  if (!ret) {
    // no support for this context type
    DebugMsg("No support for context type requested");
    return false;
  }
  // could be:
 //  WTI_INTERFACE
  //  WTI_STATUS
 //  WTI_DEFCONTEXT
  //  WTI_DEVICES
 //  WTI_CURSORS
  //  WTI_EXTENSIONS
 //  WTI_DDCTXS
  //  WTI_DSCTXS

  lcMine.lcPktData = FULLPACKETDATA;
  lcMine.lcPktMode = FULLPACKETMODE;
  lcMine.lcMoveMask = 0;
  lcMine.lcBtnUpMask = lcMine.lcBtnDnMask;

  m_data->tabCtx = pfWTOpen( m_data->hwnd, &lcMine, TRUE );

  DebugMsg("Report rate: %d Hz\n", lcMine.lcPktRate);

  // look up scaling
  {
    AXIS axes[3];
    UINT wDevice;

    /* get the data */
    //pfWTInfo(contextUsed, CTX_DEVICE, &wDevice);
    wDevice = lcMine.lcDevice;
    // pfWTInfo(WTI_DEVICES+wDevice, DVC_X, &axes[0]);
    // pfWTInfo(WTI_DEVICES+wDevice, DVC_Y, &axes[1]);
    /*
    m_data->MaxX = axes[0].axMax;
    m_data->MaxY = axes[1].axMax;
    m_data->AspectRatio = m_data->MaxY/m_data->MaxX;
    */

    pfWTInfo(WTI_DEVICES+wDevice, DVC_NPRESSURE, &axes[2]);
    m_data->scalePress = 1.0f / axes[2].axMax;
    pfWTInfo(WTI_DEVICES+wDevice, DVC_TPRESSURE, &axes[2]);
    m_data->scaleTPress = 1.0f / axes[2].axMax;

    LONG aspect = 0;
    float faspect = 1;
    pfWTInfo(contextUsed, CTX_OUTEXTY, &aspect);
    faspect = float(aspect);
    m_data->scaleY = 1.0f/aspect;
    pfWTInfo(contextUsed, CTX_OUTEXTX, &aspect);
    faspect /= float(aspect);
    m_data->scaleX = 1.0f/aspect;
    m_data->AspectRatio = faspect;

    m_data->scaleAng = 0;
    if (pfWTInfo(WTI_DEVICES+wDevice, DVC_ORIENTATION, axes))
    {
      if (axes[1].axResolution != 0)
        // axis1 seems to have valid data even if other axes don't 
        m_data->scaleAng = 2.0f *M_PIf / fix2float(axes[1].axResolution);
    }
  }
  return true;
}

bool bbTabletDevice::isValid() const
{
  return (m_data && m_data->tabCtx);
}
void bbTabletDevice::setDropOldEvents(bool onOff)
{
  if (m_data) m_data->throwAwayOldPackets = onOff;
}
bool bbTabletDevice::getDropOldEvents() const
{
  return m_data ? m_data->throwAwayOldPackets : false;
}
long bbTabletDevice::getNumDroppedEvents() const
{
  return m_data ? m_data->lostPackets : 0;
}
void bbTabletDevice::resetNumDroppedEvents()
{
  if (m_data) m_data->lostPackets = 0;
}

int bbTabletDevice::pointerMode( ) const
{
  LOGCONTEXT lctx;
  pfWTGet(m_data->tabCtx, &lctx);
  return (lctx.lcOptions & CXO_SYSTEM) ? SYSTEM_POINTER : SEPARATE_POINTER ;
}


bool bbTabletDevice::waitForNextEvent(bbTabletEvent &tev, unsigned int msec)
{
  if (!pfWTQueuePacketsEx) return false; // not inited?
  UINT FAR oldSN, newSN;
  struct timeb stamp, newstamp;
  ftime(&stamp);
  unsigned int elapsed = 0;
  while (! pfWTQueuePacketsEx(m_data->tabCtx, &oldSN, &newSN) )
  {
    ftime(&newstamp);
    elapsed = 1000 * (newstamp.time - stamp.time);
    elapsed += newstamp.millitm - stamp.millitm;
    if (elapsed > msec) return 0;
  }
  return getNextEvent(tev);
}

bool bbTabletDevice::getNextEvent(bbTabletEvent &tev)
{
  if (!pfWTPacketsGet) return false; // maybe not initialized yet
  assert(tev.rawDataSz >= sizeof(FULLPACKET));
  FULLPACKET &pkt = *(FULLPACKET*)tev.rawData;

  int got = 0;
  if (m_data->throwAwayOldPackets) {
    while (pfWTPacketsGet(m_data->tabCtx, 1, &pkt)) {
      // throwing away packets that build up!! (not such a hot idea...)
      m_data->lostPackets++;
      got++;
    }
  }
  else {
    UINT FAR oldSN, newSN;
    if (!pfWTQueuePacketsEx(m_data->tabCtx, &oldSN, &newSN)) return false;
    got = pfWTPacketsGet(m_data->tabCtx, 1, &pkt);
  }

  // in the test program, this doesn't trigger very often.
  //if (got>1) DebugMsg("-----!!!!  Tossed %d messages ----------\n", got-1);
  
  if (got && pkt.pkChanged)
  {
    //bbTabletEvent *tev = new bbTabletEvent();
#if 0
    DebugMsg("PKChanges: ");
    if (pkt.pkChanged & PK_X)                { DebugMsg("x|"); }
    if (pkt.pkChanged & PK_Y)                { DebugMsg("y|"); }
    if (pkt.pkChanged & PK_Z)                { DebugMsg("z|"); }
    if (pkt.pkChanged & PK_BUTTONS)          { DebugMsg("b|"); }
    if (pkt.pkChanged & PK_NORMAL_PRESSURE)  { DebugMsg("np|"); }
    if (pkt.pkChanged & PK_TANGENT_PRESSURE) { DebugMsg("tp|"); }
    if (pkt.pkChanged & PK_ORIENTATION)      { DebugMsg("or|"); }
    if (pkt.pkChanged & PK_ROTATION)         { DebugMsg("rot|"); }
    if (pkt.pkChanged & PK_CONTEXT)          { DebugMsg("c|"); }
    if (pkt.pkChanged & PK_STATUS)           { DebugMsg("s|"); }
    if (pkt.pkChanged & PK_TIME)             { DebugMsg("t|"); }
    if (pkt.pkChanged & PK_SERIAL_NUMBER)    { DebugMsg("sn|"); }
    if (pkt.pkChanged & PK_CHANGED)          { DebugMsg("chng|"); }
    if (pkt.pkChanged & PK_CURSOR)           { DebugMsg("csr|"); }
    DebugMsg("\n");
#endif

    // Scale data to [0,1]
    tev.x = pkt.pkX * m_data->scaleX;
    tev.y = pkt.pkY * m_data->scaleY;
    tev.z = pkt.pkZ * m_data->scaleZ;
    tev.pressure = pkt.pkNormalPressure * m_data->scalePress;
    tev.tpressure = pkt.pkTangentPressure * m_data->scaleTPress;
    tev.id = 0; // in case feature isn't supported
    pfWTInfo(WTI_CURSORS + pkt.pkCursor , CSR_PHYSID, &tev.id );
    tev.type = pkt.pkCursor;
    tev.buttons = pkt.pkButtons;

    // Now orientation is done on-demand by calling getOrientation,
    // because it's a lot of wasted computation if you don't actually want it.
    //if (pkt.pkChanged & PK_ORIENTATION)
    tev._flags &= ~BBTEV_ORIENT_OK;
    //  convertOrientation2AngleAxis(pkt.pkOrientation, m_data->scaleAng,
    //                               &tev.angle, tev.axis);
    return true;
  }
  
  return false;

}

const char *bbTabletDevice::getCursorName(int csrtype)
{
  if (csrtype < 0 || csrtype >= (int)m_data->cursors.size()) {
    return 0;
  }
  return m_data->cursors[csrtype].name.c_str();
}
int bbTabletDevice::getCursorNumButtons(int csrtype)
{
  if (csrtype < 0 || csrtype >= (int)m_data->cursors.size()) {
    return 0;
  }
  return m_data->cursors[csrtype].nButtons;
}
const char *bbTabletDevice::getCursorButtonName(int csrtype, int buttonnum)
{
  if (csrtype < 0 || csrtype >= (int)m_data->cursors.size() || buttonnum < 0) {
    return 0;
  }
  if (buttonnum >= (int)m_data->cursors[csrtype].buttonNames.size()) {
    return 0;
  }
  return m_data->cursors[csrtype].buttonNames[buttonnum].c_str();

}

//================== bbTabletEvent ====================================
bbTabletEvent::bbTabletEvent() : _flags(0) {
  rawDataSz = sizeof(FULLPACKET);
  rawData = malloc(rawDataSz);
}
bbTabletEvent::~bbTabletEvent() {
  if (rawData) free(rawData);
  rawData = 0;
  rawDataSz = 0;
}
void bbTabletEvent::getOrientation(float &angle, float axis[3]) const
{
  if (!(_flags & BBTEV_ORIENT_OK)) 
  {
    FULLPACKET &pkt = *(FULLPACKET*)rawData;
    //if (pkt.pkChanged & PK_ORIENTATION)
      bbTabletDevice::getInstance().m_data->_convertOrientation2AngleAxis(
        pkt.pkOrientation, &_angle, _axis);
    _flags |= BBTEV_ORIENT_OK;
  }
  angle = _angle;
  memcpy(axis,_axis,3*sizeof(_axis[0]));
}

const char *bbTabletEvent::getCursorName() const
{
  return bbTabletDevice::getInstance().getCursorName(type);
}

int bbTabletEvent::getCursorNumButtons() const
{
  return bbTabletDevice::getInstance().getCursorNumButtons(type);
}

const char *bbTabletEvent::getCursorButtonName(int buttonnum) const
{
  return bbTabletDevice::getInstance().getCursorButtonName(type,buttonnum);
}

int bbTabletEvent::getRawX() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkX;
}
int bbTabletEvent::getRawY() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkY;
}
int bbTabletEvent::getRawZ() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkZ;
}
int bbTabletEvent::getRawPressure() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkNormalPressure;
}
int bbTabletEvent::getRawTangentialPressure() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkTangentPressure;
}
int bbTabletEvent::getRawOrientAzimuth() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkOrientation.orAzimuth;
}
int bbTabletEvent::getRawOrientAltitude() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkOrientation.orAltitude;
}
int bbTabletEvent::getRawOrientTwist() const 
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkOrientation.orTwist;
}
int bbTabletEvent::getRawTiltX() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkTilt.tiltX;
}
int bbTabletEvent::getRawTiltY() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkTilt.tiltY;
}
int bbTabletEvent::getRawRoll() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkRotation.roRoll;
}
int bbTabletEvent::getRawPitch() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkRotation.roPitch;
}
int bbTabletEvent::getRawYaw() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkRotation.roYaw;
}
