//============================================================================
// Bare-Bones Tablet Interface for C++
//
//  This header was designed to be platform-neutral for portability, 
//  but currently Win32 is the only implemented back end.
//
//  Relies on Wintab32.dll for windows operation.
//  If you have a tablet, then you should have a copy of Wintab32.dll that
//  was installed by the tablet's driver installer.
//  
//  Also necessary (for development on Windows) will be the WINTAB SDK.
//  It can be downloaded for free from http://www.lcs-telegraphics.com/
//  It contains the necessary headers: wintab.h, pktdef.h ...
//
//  Author:        William Baxter (baxter@cs.unc.edu)
//  Last Modified: Aug 2002
//============================================================================
//  Copyright 2002 
//        William Baxter and
//        The University of North Carolina at Chapel Hill
// 
//  Permission to use, copy, modify, distribute and sell this software
//  and its documentation for any purpose is hereby granted without
//  fee, provided that the above copyright notice appear in all copies
//  and that both that copyright notice and this permission notice
//  appear in supporting documentation.  Binaries may be compiled with
//  this software without any royalties or restrictions.
// 
//  The University of North Carolina at Chapel Hill makes no
//  representations about the suitability of this software for any
//  purpose. It is provided "as is" without express or implied
//  warranty.
//============================================================================


#ifndef BB_TABLET_HEADER
#define BB_TABLET_HEADER

#define BBTAB_NOAUTOLINK
// autolinking causes troubles because we can't reliably detect when the
// app is linking against the DLL Runtime Libs.  The ifdef _DLL below doesn't
// do the trick.
#if defined(_WIN32) && !defined(BBTAB_NOAUTOLINK)
#ifdef _DLL
 // Decide what version of bbtablet.lib to link with and auto-link it.
 #ifdef _DEBUG
 #pragma comment (lib, "bbtabletds.lib") // for /MDd apps
 #else
 #pragma comment (lib, "bbtablets.lib") // for /MD apps
 #endif
#else 
 #ifdef _DEBUG
 #pragma comment (lib, "bbtabletd.lib") // for /MTd, /MLd apps
 #else 
 #pragma comment (lib, "bbtablet.lib") // for /ML, /MT apps
 #endif
#endif
#endif


/**
 * @class bbTabletEvent
 */
class bbTabletEvent
{
public:
  float x, y, z;    // Normalized [0,1] x,y,z coords 
  float pressure;   // Normalized [0,1] pressure
  float tpressure;  // "Tangential pressure" like from airbrush wheel
  unsigned int id;      // Unique physical id of stylus if supported
  unsigned int buttons; // Bitmask of pressed buttons
  int type; // The type of stylus reported by WinTab

  void getOrientation(float &angle, float axis[3]) const;
  const char *getCursorName() const;
  int getCursorNumButtons() const;
  const char *getCursorButtonName(int buttonnum) const;
  
  // Routines to get the raw values reported by the driver
  int getRawX() const;
  int getRawY() const;
  int getRawZ() const;
  int getRawPressure() const;
  int getRawTangentialPressure() const;
  int getRawOrientAzimuth() const;
  int getRawOrientAltitude() const;
  int getRawOrientTwist() const;
  int getRawTiltX() const;
  int getRawTiltY() const;
  int getRawRoll() const; // does anything support this?
  int getRawPitch() const; // does anything support this?
  int getRawYaw() const; // does anything support this?

  void *rawData;
  size_t rawDataSz;

public:
  bbTabletEvent();
  ~bbTabletEvent();
private:
  unsigned int _flags;
  mutable float _angle;
  mutable float _axis[3];
  friend class bbTabletDevice;
};


class _bbTabletDeviceInternal;

/**
 * @class bbTabletDevice
 *
 *   This class is a singleton. If multiple simultaneous tablets are supported
 *   by the drivers, this will be able get all their events.
 */
class bbTabletDevice 
{
private:
  // bbTabletDevice is a singleton, so it can't be user-instantiated or copied
  // Use getInstance() instead.
  bbTabletDevice(); 
  bbTabletDevice(const bbTabletDevice&);

public:
  enum { SYSTEM_POINTER, SEPARATE_POINTER };

  static bbTabletDevice &getInstance();

  /** 
   * initTablet
   * @param system_window  should be an HWND on Windows, e.g.
   * @param mode 
   *      SYSTEM_POINTER if this tablet is to drive  the system cursor
   *      SEPARATE_POINTER if this tablet doesn't affect the system cursor
   * @return true if successfully initialized
   */
  bool initTablet( void *system_window, int mode = SYSTEM_POINTER );

  /**
   * setPointerMode
   * @param mode
   *      SYSTEM_POINTER if this tablet is to drive  the system cursor
   *      SEPARATE_POINTER if this tablet doesn't affect the system cursor
   * @return true if pointer mode successfully set
   */
  bool setPointerMode( int mode );
  /**
   * pointerMode
   * Returns the current pointer mode, either SYSTEM_PONITER or SEPARATE_POINTER
   */
  int pointerMode( ) const;
  /**
   * isValid
   *    Returns true if the tablet has been successfully initialized
   */
  bool isValid() const;

  /**
   * setDropOldEvents
   *    Sets whether getNextEvent will drop events if called when there is
   *    more than one event waiting in the queue.  Off by default.  Turning this on is not 
   *    recommended if you have multiple tablets or plan to use multiple 
   *    styluses simultaneously.
   */
  void setDropOldEvents(bool onOff);
  bool getDropOldEvents() const;

  /**
   * getNumDroppedEvents
   *     If setDropOldEvents() is on, then this will return the number of
   *     events that were dropped.
   */
  long getNumDroppedEvents() const;
  void resetNumDroppedEvents();
  
  bool getNextEvent(bbTabletEvent &evt);
  bool waitForNextEvent(bbTabletEvent &evt, unsigned int msec = 0);

  const char *getCursorName(int csrtype);
  int getCursorNumButtons(int csrtype);
  const char *getCursorButtonName(int csrtype, int buttonnum);

private:
  _bbTabletDeviceInternal *m_data;
  friend class bbTabletEvent;
};



#endif /* BB_TABLET_HEADER */
